

//#include <stdlib.h>
//#include <iostream.h>
//#include <stdarg.h>
extern "C" {
  int printf(char*, ...) ;
}
extern "C" {
  #include "mpi.h"
}
#include "Seq.h"
#include "Procs.h"
#include "Range.h"
#include "Array.h"
#include "Threads.h"
#include "ProcTree.h"
#include "reduce.h"


void redarr(int rank, int* exts,
	    char* acc, int* accStrs,
	    char* val, int* valStrs,
	    int vecExt, int vecStr,
	    void (*opAccum)(void*, void*)) {

  // Accumulating reduction of a homogenous array of vectors.

  if(rank == 0)
    for(int i = 0 ; i < vecExt ; i++) {
      opAccum(acc, val) ;
      val += vecStr ;
    }
  else
    for(int i = 0 ; i < *exts ; i++) {
      redarr(rank - 1, exts + 1,
             acc, accStrs + 1, val, valStrs + 1,
             vecExt, vecStr, opAccum) ;

      acc += *accStrs ;
      val += *valStrs ;
    }
}

void red2arr(int rank, int* exts,
	     char* acc, int* accStrs,
	     char* val1, int* val1Strs, char* val2, int* val2Strs,
	     int vecExt, int vec1Str, int vec2Str,
	     void (*opMulAccum)(void*, void*, void*)) {

  // Accumulating binary reduction of two homogenous arrays of vectors.

  if(rank == 0)
    for(int i = 0 ; i < vecExt ; i++) {
      opMulAccum(acc, val1, val2) ;
      val1 += vec1Str ;
      val2 += vec2Str ;
    }
  else
    for(int i = 0 ; i < *exts ; i++) {
      red2arr(rank - 1, exts + 1,
              acc, accStrs + 1,
	      val1, val1Strs + 1, val2, val2Strs + 1,
              vecExt, vec1Str, vec2Str,
              opMulAccum) ;

      acc += *accStrs ;
      val1 += *val1Strs ;
      val2 += *val2Strs ;
    }
}

void maparr(int rank, int* exts,
            char* arg, int* argStrs,
            void (*op)(void*)) {

  // Simple procedural `map' operation on an array.

  if(rank == 0)
    (*op)(arg) ;
  else
    for(int i = 0 ; i < *exts ; i++) {
      maparr(rank - 1, exts + 1, arg, argStrs + 1, op) ;

      arg += *argStrs ;
    }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Scalar reduction operation.                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


static void reduceLoop(char* resData, char* srcData,
		       Frame* src,
		       ProcTree* reductionTree,
		       const int d,
		       void (*opAccum)(void*, void*)) {

  int rank = src->rnk() ;

  if(d == rank)
    (*opAccum)(resData, srcData) ;
  else {
    Range* x = src->rng(d + 1) ;

    Stride* srcStr = src->str(d + 1) ;

    for(x->locBlocksEnum() ; x->locBlocksTest() ; x->locBlocksNext()) {
      char* srcDataElem = srcData + srcStr->offset(x) ;
      int step  = srcStr->step(x) ;
      int count = x->blockCount() ;

      for(int i = 0 ; i < count ; i++) {
	reduceLoop(resData, srcDataElem, src, reductionTree, d + 1, opAccum) ;

        srcDataElem += step ;
      }
    }
  }
}


void reduce(const int typeLen, char* res, Frame* src,
	    void (*opZero)(void*), void (*opAccum)(void*, void*)) {

  ProcessGroup* p = src->prc() ;
  Data resFrameP(typeLen, p, 0, NULL) ;

  if(p->on()) {
    ProcTree reductionTree ;

    ProcessGroup q(*p) ;
    for(int dim = 1 ; dim <= src->rnk() ; dim++) {
      Range* x = src->rng(dim) ;
      Embed e(0, *x) ;
      if(x->distributed()) q.deactivate(&e) ;
    }
    apg->radialTree(&reductionTree, &q) ;

    char* resData = resFrameP.dat() ;

    opZero(resData) ;
    reduceLoop(resData, src->dat(), src,
               &reductionTree, 0, opAccum) ;

    reductionTree.red(typeLen, resData, opAccum) ;
  } p->no() ;

  // Broadcast result to active process group
  ProcessGroup glob(*apg) ;
  Promotion resFrame(typeLen, &glob, 0, NULL, res) ;
  remap(&resFrame, &resFrameP) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Scalar binary reduction operation (dot product).                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


static void reduce2Loop(char* resData, char* arg1Data, char* arg2Data,
		        Frame* arg1, Frame* arg2,
		        ProcTree* reductionTree,
		        const int d,
		        void (*opMulAccum)(void*, void*, void*)) {

  int rank = arg1->rnk() ;

  if(d == rank)
    (*opMulAccum)(resData, arg1Data, arg2Data) ;
  else {
    Range* x = arg1->rng(d + 1) ;

    Stride* arg1Str = arg1->str(d + 1) ;
    Stride* arg2Str = arg2->str(d + 1) ;

    for(x->locBlocksEnum() ; x->locBlocksTest() ; x->locBlocksNext()) {
      char* arg1DataElem = arg1Data + arg1Str->offset(x) ;
      char* arg2DataElem = arg2Data + arg2Str->offset(x) ;
      int step1  = arg1Str->step(x) ;
      int step2  = arg2Str->step(x) ;
      int count = x->blockCount() ;

      for(int i = 0 ; i < count ; i++) {
	reduce2Loop(resData, arg1DataElem, arg2DataElem, arg1, arg2,
                    reductionTree, d + 1, opMulAccum) ;

        arg1DataElem += step1 ;
        arg2DataElem += step2 ;
      }
    }
  }
}


void reduce2(const int typeLen, char* res, Frame* arg1, Frame* arg2,
	     void (*opZero)(void*),
             void (*opMulAccum)(void*, void*, void*),
             void (*opAccum)(void*, void*)) {

  // `arg1' and `arg2' must have the same shape and mapping.

  ProcessGroup* p = arg1->prc() ;
  Data resFrameP(typeLen, p, 0, NULL) ;

  if(p->on()) {
    ProcTree reductionTree ;

    ProcessGroup q(*p) ;
    for(int dim = 1 ; dim <= arg1->rnk() ; dim++) {
      Range* x = arg1->rng(dim) ;
      Embed e(0, *x) ;
      if(x->distributed()) q.deactivate(&e) ;
    }
    apg->radialTree(&reductionTree, &q) ;

    char* resData = resFrameP.dat() ;

    opZero(resData) ;
    reduce2Loop(resData, arg1->dat(), arg2->dat(), arg1, arg2,
                &reductionTree, 0, opMulAccum) ;

    reductionTree.red(typeLen, resData, opAccum) ;
  } p->no() ;

  // Broadcast result to active process group
  ProcessGroup glob(*apg) ;
  Promotion resFrame(typeLen, &glob, 0, NULL, res) ;
  remap(&resFrame, &resFrameP) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Array-valued dimensional reduction operation.                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


static void reduceDimLoop(char* resData, char* srcData,
			  int* exts, int* srcSteps, int* resSteps,
			  const int d, const int r,
			  Frame* res, Frame* src,
			  const int dim, ProcTree* reductionTree,
			  void (*opZero)(void*),
                          void (*opAccum)(void*, void*)) {

  int rank = src->rnk() ;

  if(d == rank) {
    Range* x = src->rng(dim) ;

    Stride* srcStr = src->str(dim) ;

    maparr(rank - 1, exts, resData, resSteps, opZero) ;
    for(x->locBlocksEnum() ; x->locBlocksTest() ; x->locBlocksNext())
      redarr(rank - 1, exts,
	     resData, resSteps, srcData + srcStr->offset(x), srcSteps,
	     x->blockCount(), srcStr->step(x), opAccum) ;

    reductionTree->redArray(src->len(), rank - 1, exts,
                            resData, resSteps, opAccum) ;
  }
  else
    if(d == dim - 1)
      reduceDimLoop(resData, srcData, exts, srcSteps, resSteps, d + 1, r,
                    res, src, dim, reductionTree, opZero, opAccum) ;
    else {
      Range* x = src->rng(d + 1) ;

      Stride* srcStr = src->str(d + 1) ;
      Stride* resStr = res->str(r + 1) ;

      for(x->locBlocksEnum() ; x->locBlocksTest() ; x->locBlocksNext()) {
        char* srcDataElem = srcData + srcStr->offset(x) ;
        char* resDataElem = resData + resStr->offset(x) ;

        exts [r] = x->blockCount() ;
        srcSteps [r] = srcStr->step(x) ;
        resSteps [r] = resStr->step(x) ;
 
        reduceDimLoop(resDataElem, srcDataElem, exts, srcSteps, resSteps,
                      d + 1, r + 1,
                      res, src, dim, reductionTree, opZero, opAccum) ;
      }
    }
}


void reduceDim(Frame* res, Frame* src, const int dim, 
	       void (*opZero)(void*), void (*opAccum)(void*, void*)) {

  if(src->prc()->on()) {
    ProcTree reductionTree ;

    ProcessGroup q(*(src->prc())) ;
    Range* x = src->rng(dim) ;
    Embed e(0, *x) ;
    if(x->distributed()) q.deactivate(&e) ;
    apg->radialTree(&reductionTree, &q) ;

    int* exts = new int [res->rnk()] ;
    int* srcSteps = new int [res->rnk()] ;
    int* resSteps = new int [res->rnk()] ;

    reduceDimLoop(res->dat(), src->dat(), exts, srcSteps, resSteps, 0, 0,
                  res, src, dim, &reductionTree, opZero, opAccum) ;

    delete resSteps ;
    delete srcSteps ;
    delete exts ;
  } src->prc()->no() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Array-valued binary reduction operation (dot product).             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


static void reduce2DimLoop(char* resData, char* arg1Data, char* arg2Data,
			   int* exts,
                           int* resSteps, int* arg1Steps, int* arg2Steps,
			   const int d, const int r,
			   Frame* res, Frame* arg1, Frame* arg2,
			   const int dim, ProcTree* reductionTree,
			   void (*opZero)(void*),
		           void (*opMulAccum)(void*, void*, void*),
                           void (*opAccum)(void*, void*)) {

  int rank = arg1->rnk() ;

  if(d == rank) {
    Range* x = arg1->rng(dim) ;

    Stride* arg1Str = arg1->str(dim) ;
    Stride* arg2Str = arg2->str(dim) ;

    maparr(rank - 1, exts, resData, resSteps, opZero) ;
    for(x->locBlocksEnum() ; x->locBlocksTest() ; x->locBlocksNext())
      red2arr(rank - 1, exts,
	      resData, resSteps,
              arg1Data + arg1Str->offset(x), arg1Steps,
              arg2Data + arg2Str->offset(x), arg2Steps,
	      x->blockCount(), arg1Str->step(x), arg2Str->step(x),
              opMulAccum) ;

    reductionTree->redArray(arg1->len(), rank - 1, exts,
                            resData, resSteps, opAccum) ;
  }
  else
    if(d == dim - 1)
      reduce2DimLoop(resData, arg1Data, arg2Data,
                     exts, resSteps, arg1Steps, arg2Steps, d + 1, r,
                     res, arg1, arg2, dim, reductionTree,
                     opZero, opMulAccum, opAccum) ;
    else {
      Range* x = arg1->rng(d + 1) ;

      Stride* arg1Str = arg1->str(d + 1) ;
      Stride* arg2Str = arg2->str(d + 1) ;
      Stride* resStr = res->str(r + 1) ;

      for(x->locBlocksEnum() ; x->locBlocksTest() ; x->locBlocksNext()) {
        char* arg1DataElem = arg1Data + arg1Str->offset(x) ;
        char* arg2DataElem = arg2Data + arg2Str->offset(x) ;
        char* resDataElem = resData + resStr->offset(x) ;

        exts [r] = x->blockCount() ;
        arg1Steps [r] = arg1Str->step(x) ;
        arg2Steps [r] = arg2Str->step(x) ;
        resSteps [r] = resStr->step(x) ;
 
        reduce2DimLoop(resDataElem, arg1DataElem, arg2DataElem,
                       exts, resSteps, arg1Steps, arg2Steps, d + 1, r + 1,
                       res, arg1, arg2, dim, reductionTree,
                       opZero, opMulAccum, opAccum) ;
      }
    }
}


void reduce2Dim(Frame* res, Frame* arg1, Frame* arg2, const int dim, 
	        void (*opZero)(void*),
	        void (*opMulAccum)(void*, void*, void*),
	        void (*opAccum)(void*, void*)) {

  ProcessGroup* p = arg1->prc() ;

  if(p->on()) {
    ProcTree reductionTree ;

    {
      ProcessGroup q(*p) ;
      Range* x = arg1->rng(dim) ;
      Embed e(0, *x) ;
      if(x->distributed()) q.deactivate(&e) ;
      apg->radialTree(&reductionTree, &q) ;
    }

    int* exts = new int [res->rnk()] ;
    int* arg1Steps = new int [res->rnk()] ;
    int* arg2Steps = new int [res->rnk()] ;
    int* resSteps = new int [res->rnk()] ;

    reduce2DimLoop(res->dat(), arg1->dat(), arg2->dat(),
                   exts, resSteps, arg1Steps, arg2Steps, 0, 0,
                   res, arg1, arg2, dim, &reductionTree,
                   opZero, opMulAccum, opAccum) ;

    delete resSteps ;
    delete arg2Steps ;
    delete arg1Steps ;
    delete exts ;
  } p->no() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Location of maximum element.                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


static void maxLocLoop(char* resData, int* loc, char* srcData,
                       int* idx,
		       Frame* src,
		       ProcTree* reductionTree,
		       const int d,
		       int (*opAccumMax)(void*, void*)) {

  int rank = src->rnk() ;

  if(d == rank) {
    if((*opAccumMax)(resData, srcData))
      for(int i = 0 ; i < rank ; i++)
        loc [i] = idx [i] + 1 ;       // Fortran subscripts!
  }
  else {
    Range* x = src->rng(d + 1) ;

    Stride* srcStr = src->str(d + 1) ;

    for(x->locBlocksEnum() ; x->locBlocksTest() ; x->locBlocksNext()) {
      char* srcDataElem = srcData + srcStr->offset(x) ;
      int step  = srcStr->step(x) ;
      int count = x->blockCount() ;

      idx [d] = (*x)() ;

      for(int i = 0 ; i < count ; i++) {
	maxLocLoop(resData, loc, srcDataElem, idx, src,
                   reductionTree, d + 1, opAccumMax) ;

        srcDataElem += step ;
        (idx [d])++ ;
      }
    }
  }
}


void maxLoc(int typeLen, char* res, int* loc, Frame* src,
	    void (*opLoValue)(void*), int (*opAccumMax)(void*, void*)) {

  const int locExt = src->rnk() ;
  const int locLen = locExt * sizeof(int) ;

  ProcessGroup* p = src->prc() ;
  Data resFrameP(typeLen, p, 0, NULL) ;
  Data locFrameP(locLen, p, 0, NULL) ;

  if(p->on()) {
    ProcTree reductionTree ;

    ProcessGroup q(*p) ;
    for(int dim = 1 ; dim <= src->rnk() ; dim++) {
      Range* x = src->rng(dim) ;
      Embed e(0, *x) ;
      if(x->distributed()) q.deactivate(&e) ;
    }
    apg->radialTree(&reductionTree, &q) ;

    char* resData = resFrameP.dat() ;
    int* locData  = (int*) locFrameP.dat() ;

    opLoValue(resData) ;
    int* idx = new int [locExt] ;
    maxLocLoop(resData, locData, src->dat(), idx,
               src, &reductionTree, 0, opAccumMax) ;
    delete idx ;

    reductionTree.redTag(typeLen, resData, locLen, (char*) locData,
                         opAccumMax) ;
  } p->no() ;

  // Broadcast result to active process group
  ProcessGroup glob(*apg) ;
  Promotion resFrame(typeLen, &glob, 0, NULL, res) ;
  Promotion locFrame(locLen, &glob, 0, NULL, (char*) loc) ;
  remap(&resFrame, &resFrameP) ;
  remap(&locFrame, &locFrameP) ;
}


// ----------------------------------------------------------------------
// Things to do
//
//      Change the `red' primitives so that they reduce only, and
//      perform broadcast by subsequent `remap'?

